/* $Id: cgistream.c,v 1.8 2002/06/14 13:54:04 aa5779 Exp $ */
/*
	Copyright (C) 2001 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "util.h"
#include "lists.h"
#include "interp.h"
#define CGISTREAM_IMP
#include "cgistream.h"
#undef putc

stream_class incgi_stream_class = {"incgi", ics_open, ics_close, ics_read, 
									dummy_writestr, 
								    dummy_writetag, dummy_ioctl, dummy_getpos,
								    dummy_setpos, dummy_rewind
								 };

stream_class outcgi_stream_class = {"outcgi", ocs_open, hs_close, dummy_read,
									ocs_writestr, dummy_writetag, dummy_ioctl,
									dummy_getpos, dummy_setpos, dummy_rewind,
									};
									
static stream *ics_stream;
static html_chunk form_chunk;

static char *strtok_x(char *data, char *delims)
{
	static char *next;
	char *ptr;
	char *base;

	if(data)
		next = data;
	if(!next)
		return NULL;
	base = next;
	ptr = next + strcspn(next, delims);
	next = *ptr ? ptr + 1 : NULL;
	*ptr = '\0';
	return base;
}

static void parse_url_encode(char *buffer)
{
	char *name = strtok_x(buffer, "=");
	char *value;
	dstring *val;
	while(name)
	{
		val = ds_create(NULL);
		value = strtok_x(NULL, "&;");
		if(!value)
			break;
		for(; *value; value++)
		{
			switch(*value)
			{
				case '+':
					ds_appendch(val, ' ');
					break;
				case '%':
				{
					int ch = toupper(*++value);
					int v = isalpha(ch) ? ch - 'A' + 10 : ch - '0';
					ch = toupper(*++value);
					v = (v << 4) | (isalpha(ch) ? ch - 'A' + 10 : ch - '0');
					ds_appendch(val, v);
					break;
				}
				default:
					ds_appendch(val, *value);
			}
		}
		set_html_param(&form_chunk, ds_create(name), val);
		name = strtok_x(NULL, "=");
	}
}

stream *ics_open(interpreter *i, char *name, char *fname, int writeable)
{
	char *req_method = getenv("REQUEST_METHOD");
	char *query;
	if(ics_stream)
		fatal_error("Cannot use multiple CGI streams");
	ics_stream = ALLOC(stream);
	ics_stream->class = &incgi_stream_class;
	ics_stream->direction = stream_in;
	if(!req_method || !strcmp(req_method, "GET"))
	{
		query = getenv("QUERY_STRING");
		query = xstrdup(query ? query : "");
	}
	else if(!strcmp(req_method, "POST"))
	{
		int cnt = atoi(getenv("CONTENT_LENGTH"));
		query = xmalloc(cnt + 1);
		read(STDIN_FILENO, query, cnt);
		query[cnt] = '\0';
		if(strcmp(getenv("CONTENT_TYPE"), "application/x-www-form-urlencoded"))
		{
			set_tag_name(&form_chunk, ds_create("QUERY"));
			set_html_param(&form_chunk, ds_create("rawdata"), ds_create(query));
			free(query);
			return ics_stream;
		}
	}
	else query = xstrdup("");
	set_tag_name(&form_chunk, ds_create("QUERY"));
	parse_url_encode(query);
	free(query);
	return ics_stream;
}

void ics_close(interpreter *i, stream *s)
{
	ics_stream = NULL;
}

html_chunk *ics_read(interpreter *i, stream *s)
{
	return &form_chunk;
}

stream *ocs_open(interpreter *i, char *name, char *fname, int writeable)
{
	stream *s = hs_open(i, name, "", 1);
	s->class = &outcgi_stream_class;
	return s;
}
	

void ocs_writestr(interpreter *i, stream *s, char *str)
{
	html_stream *hs = s->data;
	if(*str)
		hs->basic.fc->puts(hs->basic.fh, str);
	else s->class = &html_stream_class;
	hs->basic.fc->putc(hs->basic.fh, '\n');
}



