/*$Id: shared.c,v 1.9 2002/06/14 13:54:04 aa5779 Exp $*/
/*
	Copyright (C) 2001 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <string.h>
#include <gfnmatch.h>
#include "lists.h"
#include "util.h"
#include "dstring.h"
#include "interp.h"

extern shared_object_class *get_so_class(char *cls_name);

int find_byref(void *obj, void *key)
{
	return obj == key;
}

static int find_in_names(exported_names *en, char *key)
{
	return !strcmp(en->extname, key);
}

static void destroy_extname(exported_names *en)
{
	free(en->extname);
}

int so_adjust;
int so_find(void *obj, char *name)
{
	shared_object *so = (shared_object *)((char *)obj + so_adjust);
	return list_find(so->extnames, name, (list_predicate_t)find_in_names) != NULL;
}

int si_find_exact(shared_int *si, char *name)
{
	return !strcmp(si->intname, name);
}

int si_find_match(shared_int *si, char *key)
{
	return !fnmatch(si->intname, key, 0);
}

int si_find_byso(shared_int *si, void *so)
{
	return si->obj == so;
}

void destroy_shared_int(shared_int *si)
{
	shared_object *so = si->so;
	shared_object_class *soc = so->class;
	free(si->intname);
	if(!so->extnames || !--so->refcnt)
		*soc->base = list_remove(*soc->base, si->obj, (list_predicate_t)find_byref, soc->destroy, soc->size);
}

void destroy_so(shared_object *so)
{
	list_delete(so->extnames, (list_destroy_t)destroy_extname, sizeof(exported_names));
}

void export_shared(shared_object *so, char *extname)
{
	exported_names *en = ALLOC(exported_names);
	en->extname = xstrdup(extname);
	so->extnames = list_add(so->extnames, en);
}



void *find_exported(char *clsname, char *name)
{
	shared_object_class *cls = get_so_class(clsname);
	so_adjust = cls->offs;
	return list_find(*cls->base, name, (list_predicate_t)so_find);
}

shared_int *embed_shared(void *object, char *clsname, char *name)
{
	shared_object_class *cls = get_so_class(clsname);
	shared_int *si;
	si = ALLOC(shared_int);
	si->intname = xstrdup(name);
	si->obj = object;
	si->so = (shared_object *)((char *)object + cls->offs);
	si->so->refcnt++;
	return si;
}

shared_int *import_shared(char *iname, char *name, char *clsname)
{
	shared_object_class *cls = get_so_class(clsname);
	void *so;
	so_adjust = cls->offs;
	so = list_find(*cls->base, name, (list_predicate_t)so_find);
	if(!so)
		recover_error("can't import %s '%s'", clsname, name);
	return embed_shared(so, clsname, iname);
}


shared_int *init_shared_object(interpreter *i, void *so, char *clsname, char *intname)
{
	shared_object_class *soc = get_so_class(clsname);
	shared_object *so1;
	shared_int *si;
	
	so1 = (shared_object *)((char *)so + soc->offs);
	so1->refcnt = 1;
	so1->class = soc;
	so1->owner = i;
	*soc->base = list_add(*soc->base, so);
	si = ALLOC(shared_int);
	si->intname = xstrdup(intname);
	si->obj = so;
	si->so = so1;
	return si;
}

shared_int *find_si_itself(interpreter *i, char *cls, char *iname)
{
	shared_object_class *soc = get_so_class(cls);
	shared_int **lst = (shared_int **)((char *)i + soc->offs2);
	shared_int *si = list_find(*lst, iname, (list_predicate_t)si_find_exact);
	return si;
}

void *quiet_find_shared_int(interpreter *i, char *cls, char *iname)
{
	shared_int *si = find_si_itself(i, cls, iname);
	return si ? si->obj : NULL;
}

void *find_shared_int(interpreter *i, char *cls, char *iname)
{
	void *so = quiet_find_shared_int(i, cls, iname);
	if(!so)
		recover_error("%s '%s' undefined", cls, iname);
	return so;
}

static int al_find(allow_list *al, interpreter *key)
{
	return !al->subject || al->subject == key;
}

static int al_find_exact(allow_list *al, interpreter *key)
{
	return al->subject == key;
}

void check_si_perm(interpreter *i, shared_int *si)
{
	if(i != si->so->owner && !list_find(si->so->allowed, i, (list_predicate_t)al_find) &&
	   list_find(si->so->denied, i, (list_predicate_t)al_find))
		recover_error("permission denied to use an object");
}

void allow_shared_int(interpreter *owner, shared_int *si, interpreter *dest)
{
	allow_list *al;
	if(si->so->owner != owner)
		recover_error("only object owner may change its access lists");
	if(!dest)
	{
		list_delete(si->so->denied, NULL, sizeof(*si->so->denied));
		si->so->denied = NULL;
	}
	else
		si->so->denied = list_remove(si->so->denied, dest, (list_predicate_t)al_find_exact, NULL, sizeof(*si->so->denied));
	al = ALLOC(allow_list);
	al->subject = dest;
	si->so->allowed = list_add(si->so->allowed, al);
}

void deny_shared_int(interpreter *owner, shared_int *si, interpreter *dest)
{
	allow_list *al;
	if(si->so->owner != owner)
		recover_error("only object owner may change its access lists");
	if(!dest)
	{
		list_delete(si->so->allowed, NULL, sizeof(*si->so->allowed));
		si->so->allowed = NULL;
	}
	else
		si->so->allowed = list_remove(si->so->allowed, dest, (list_predicate_t)al_find_exact, NULL, sizeof(*si->so->denied));
	al = ALLOC(allow_list);
	al->subject = dest;
	si->so->denied = list_add(si->so->denied, al);
}

void rename_shared_int(interpreter *i, char *cls, char *name, char *newname, int force)
{
	shared_int *si = find_si_itself(i, cls, name);

	if(!si)
	{
		if(force)
			return;
		recover_error("can't rename %s %s: not defined", cls, name);
	}
	free(si->intname);
	si->intname = xstrdup(newname);
}

void add_shared_int(interpreter *i, shared_int *si)
{
	shared_int **lst = (shared_int **)((char *)i + si->so->class->offs2);
	check_si_perm(i, si);
	*lst = list_add(*lst, si);
}

